
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>

#ifndef MAIN_DEFINED
    #include "main.h"
#endif

/* global socket connection id */
int gbConnID;

/* Board height and width */
int gbWidth, gbHeight;


/* Our configuration */
int gbMyID;
int gbMyCapacity;
int gbMyMoney;
int gbMyX;
int gbMyY;

HOMEBASENODE *gbHomeBase;
RBTLISTNODE *gbRbts;
PKGLISTNODE *gbMyPkgs;
PKGLISTNODE *gbKnownPkgs;
TILE *gbBoard;


FILE *gbConnPtr;
char gbline[MAX_LINE+1];

void DoInitPhase(char *server, int port)
{
    struct sockaddr_in serverInfo;
    struct hostent *hostIP;

    // Create a client socket
    if ( (gbConnID = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        printf("Oops! Could not create client socket\n");
        exit(0);
        return;
    }

    // Resolve server name
    hostIP = gethostbyname(server);
    if (!hostIP)
    {
        printf("Oops! Could not resolve server name %s\n", server);
        exit(0);
        return;
    }

    serverInfo.sin_addr =  *((struct in_addr *)hostIP->h_addr);
/*  serverInfo.sin_addr.s_addr = inet_addr("10.212.10.25");*/
    serverInfo.sin_family = AF_INET;
    serverInfo.sin_port = htons(port);
    memset(&(serverInfo.sin_zero), '\0', 8);

    // Finally! Connect to server
    if ( (connect(gbConnID, (struct sockaddr *) &serverInfo, sizeof(struct sockaddr))) < 0)
    {
        printf("Oops! Could not connect to server %s\n", server);
        exit(0);
        return;
    }

    gbConnPtr = fdopen(gbConnID, "w+");
    if(gbConnPtr == NULL)
    {
          printf("fdopen failed. gbConnID:%d\n", gbConnID);
          exit(0);
    }
    // Send START_STR
    SendALine(START_STR);

}


void ReadBoard()
{
    int i, j;
    HOMEBASENODE *newNode;

    // Read the dimensions of board
    char *line = GetNextLine();
    sscanf(line, "%d %d", &gbWidth, &gbHeight);

#ifdef ICFP_DEBUG
    printf("ReadBoard %d %d\n", gbWidth, gbHeight);
#endif

    // Allocate space
    gbBoard = (TILE *) malloc(MEM_WIDTH * MEM_HEIGHT*sizeof(TILE));
    if(gbBoard == NULL)
    {
         printf("Allocating the board failed\n");
         exit(0);
    }

    // Border tiles with WALLs
    for (j = 0; j < MEM_HEIGHT; j ++)
        Tile(0, j).type = Tile(MEM_WIDTH - 1, j).type = WALL;

    for (i = 0; i < MEM_WIDTH; i ++)
        Tile(i, 0).type = Tile(i, MEM_HEIGHT - 1).type = WALL;

    //Rest of 'em
    for (j = 1; j <= gbHeight; j ++)
    {
        // Read row
        line = GetNextLine();
#ifdef ICFP_DEBUG
        printf("Read Row %s\n", line);
#endif
        for (i = 1; i <= gbWidth; i ++ )
        {
            sscanf(line, "%c", &(Tile(i, j).type));
            if (Tile(i, j).type == HOME)
            {
                newNode = NewHomeBaseNode(i, j);
                if(newNode)
                {
                    gbHomeBase = AddHomeBaseNode(gbHomeBase, newNode);
                }
            }
            line ++;
        }
    }
}

void ReadConfig()
{
    char *configLine = GetNextLine();
    sscanf(configLine, "%d %d %d", &gbMyID, &gbMyCapacity, &gbMyMoney);
}

void ReadSrvrResponse()
{
    char *responseLine;
    char *token;
    RBTLISTNODE *rbtNode;
    PKGLISTNODE *pkgNode;
    RBTLISTNODE *prev;
    ROBOT robot;
    PACKAGE pkg;
    int pkgID;

    rbtNode = gbRbts;
    while (rbtNode)
    {
        rbtNode->rbt.updated = 0;
        rbtNode = rbtNode->next;
    }

    responseLine = GetNextLine();
#ifdef ICFP_DEBUG
    printf("Response line = :%s:\n", responseLine);
#endif
    token = strtok(responseLine, " ");
    while (token)
    {
        if (token[0] == '#')
        {
            token ++;
            rbtNode = FindRbtNode(gbRbts, atoi(token));
            if (!rbtNode)
            {
                robot.id = atoi(token);
                robot.x = robot.y = 0; //invalid value to start with
                robot.packageIDs = NULL;
                rbtNode = NewRbtNode(&robot);
                if(rbtNode)
                {
                    goto cleanup;
                }

                gbRbts = AddRbtNode(gbRbts, rbtNode);
            }

            rbtNode->rbt.updated = 1;
            while ((token = strtok(NULL, " ")) && token[0] != '#')
            {
                switch (*token)
                {
                    case 'X' :
                        token = strtok(NULL, " ");
                        rbtNode->rbt.x = atoi(token);
                        token = strtok(NULL, " ");
                        if (*token == 'Y')
                        {
                            token = strtok(NULL, " ");
                            rbtNode->rbt.y = atoi(token);
                        }
                        if (rbtNode->rbt.id == gbMyID)
                        {
                            gbMyX = rbtNode->rbt.x;
                            gbMyY = rbtNode->rbt.y;
                        }
                        break;

                        // Robot heads eastwards
                    case 'E':
                        /*token = strtok(NULL, " ");*/
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = NULL;
                        rbtNode->rbt.x ++;
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = &(rbtNode->rbt);

                        if (rbtNode->rbt.id == gbMyID)
                        {
                            gbMyX = rbtNode->rbt.x;
                        }
                        break;

                        // And so on...
                    case 'W':
                        /*token = strtok(NULL, " ");*/
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = NULL;
                        rbtNode->rbt.x --;
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = &(rbtNode->rbt);


                        if (rbtNode->rbt.id == gbMyID)
                        {
                            gbMyX = rbtNode->rbt.x;
                        }
                        break;

                    case 'N':
                        /*token = strtok(NULL, " ");*/
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = NULL;
                        rbtNode->rbt.y ++;
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = &(rbtNode->rbt);

                        if (rbtNode->rbt.id == gbMyID)
                        {
                            gbMyY = rbtNode->rbt.y;
                        }
                        break;

                    case 'S':
                        /*token = strtok(NULL, " ");*/
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = NULL;
                        rbtNode->rbt.y --;
                        Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = &(rbtNode->rbt);

                        if (rbtNode->rbt.id == gbMyID)
                        {
                            gbMyY = rbtNode->rbt.y;
                        }
                        break;

                        // Picked a package
                    case 'P': {
                            INTLISTNODE *newInt;
                            token = strtok(NULL, " ");
                            pkgID = atoi(token);
                            newInt = NewIntNode(pkgID);

                            if(newInt)
                            {
                                rbtNode->rbt.packageIDs = AddIntNode(rbtNode->rbt.packageIDs, newInt);
                            }

                            pkgNode = FindPkgNode(gbKnownPkgs, pkgID);
                            // If this was a known package, we remove it from the known             // packages list
                            if (pkgNode)
                            {
                                if (rbtNode->rbt.id == gbMyID)
                                {
                                    PKGLISTNODE *newPkgNode = NewPkgNode(&(pkgNode->pkg));
                                    if(newPkgNode)
                                    {
                                        gbMyPkgs = AddPkgNode(gbMyPkgs, newPkgNode);
                                    }
                                }
                                gbKnownPkgs = DeletePkgNode(gbKnownPkgs, pkgID);
                            }

                            break;
                        }

                        // Dropped one
                    case 'D':
                        token = strtok(NULL, " ");
                        pkgID = atoi(token);
                        rbtNode->rbt.packageIDs = DeleteIntNode(rbtNode->rbt.packageIDs, pkgID);

                        if (rbtNode->rbt.id == gbMyID)
                        {
                            // If this wasn't the package's destination
                            // add the package to known package list

                            pkgNode = FindPkgNode(gbMyPkgs, pkgID);
                            if (pkgNode->pkg.destx != gbMyX || pkgNode->pkg.desty != gbMyY)
                            {
                                PKGLISTNODE *newPkgNode = NewPkgNode(&(pkgNode->pkg));
                                if(newPkgNode)
                                {
                                    gbKnownPkgs = AddPkgNode(gbKnownPkgs, newPkgNode);
                                }
                            }

                            gbMyPkgs = DeletePkgNode(gbMyPkgs, pkgID);
                        }

                        else
                        {
                            pkg.id = pkgID;
                            pkg.destx = pkg.desty = -1;
                            pkg.x = rbtNode->rbt.x;
                            pkg.y = rbtNode->rbt.y;
                            pkgNode = NewPkgNode(&pkg);
                            if(pkgNode)
                            {
                                gbKnownPkgs = AddPkgNode(gbKnownPkgs, pkgNode);
                            }
                        }
                        break;

                }
            }

        }
    }

cleanup:
    // Clean up dead robots
    rbtNode = gbRbts;
    prev = NULL;
    while (rbtNode)
    {
        if (rbtNode->rbt.updated == 0)
        {
            RBTLISTNODE *next = rbtNode->next;
            Tile(rbtNode->rbt.x, rbtNode->rbt.y).robot = NULL;

            FreeIntList(rbtNode->rbt.packageIDs);

            free(rbtNode);

            if (prev == NULL)
            {
                rbtNode = gbRbts = next;
            }
            else
            {
                rbtNode = prev->next = next;
            }
        }
        else
        {
            prev = rbtNode;
            rbtNode = rbtNode->next;
        }
    }
}

/*char *GetNextLine()
{
     int i = 0;
     do
     {
          recv(gbConnID, &gbline[i], 1, 0);

     }while(gbline[i++] != '\n');
     gbline[i] = '\0';
     printf("line:%s\n", gbline);
     return gbline;
}*/



char *GetNextLine()
{
    static char line[MAX_LINE+1];
    fgets(line, MAX_LINE, gbConnPtr);

    //remove newline
    line[strlen(line) - 1] = '\0';
    return line;
}

/*void SendALine(char *line)
{
     int len = strlen(line);
     if(line[len-1] != '\n')
     {
          line[len] = '\n';
          line[len+1] = '\0';
     }
     len = strlen(line);

     send(gbConnID, line, len, 0);
}*/



void SendALine(char *line)
{
#ifdef ICFP_DEBUG
    printf("Sending command %s\n", line);
#endif
    fputs(line, gbConnPtr);
}

int IntListLen(INTLISTNODE *head)
{
    int len = 0;
    while (head)
    {
        len++;
        head = head->next;
    }

    return len;
}

int PkgListLen(PKGLISTNODE *head)
{
    int len = 0;

    while (head)
    {
        len++;
        head = head->next;
    }

    return len;
}

int PkgListWtSum(PKGLISTNODE *node)
{
    int sum;

    sum = 0;

    while (node)
    {
        sum += node->pkg.weight;
        node = node->next;
    }

    return sum;
}

/* List helper functions */

INTLISTNODE *NewIntNode(int val)
{
    INTLISTNODE *newNode = (INTLISTNODE *) malloc(sizeof(INTLISTNODE));

    if(newNode)
    {
        newNode->val = val;
        newNode->next = NULL;
    }

    return newNode;
}

PKGLISTNODE *NewPkgNode(const PACKAGE *pkg)
{
    PKGLISTNODE *newNode = (PKGLISTNODE *) malloc(sizeof(PKGLISTNODE));

    if(newNode)
    {
        newNode->pkg = *pkg;
        newNode->next = NULL;
    }

    return newNode;
}

RBTLISTNODE *NewRbtNode(const ROBOT *robot)
{
    RBTLISTNODE *newNode = (RBTLISTNODE *) malloc(sizeof(RBTLISTNODE));

    if(newNode)
    {
        newNode->rbt = *robot;
        newNode->next = NULL;
    }

    return newNode;
}


INTLISTNODE *AddIntNode(INTLISTNODE *head, INTLISTNODE *newNode)
{
    if (!head)
    {
        head = newNode;
        return newNode;
    }

    newNode->next = head;
    head = newNode;
    return head;
}


PKGLISTNODE *AddPkgNode(PKGLISTNODE *head, PKGLISTNODE *newNode)
{
    if (!head)
    {
        head = newNode;
        return newNode;
    }

    newNode->next = head;
    head = newNode;
    return head;
}

RBTLISTNODE *AddRbtNode(RBTLISTNODE *head, RBTLISTNODE *newNode)
{
    if (!head)
    {
        head = newNode;
        return newNode;
    }

    newNode->next = head;
    head = newNode;
    return head;
}


INTLISTNODE *DeleteIntNode(INTLISTNODE *head, int val)
{
    INTLISTNODE *p, *q;
    if (!head)
    {
        return NULL;
    }

    if (head->val == val)
    {
        p = head;
        head = head->next;
        free(p);
        return head;
    }

    q = head;
    p = head->next;
    while (p)
    {
        if (p->val == val)
        {
            q->next = p->next;
            free(p);
            return head;
        }
        q = p;
        p = p->next;
    }
    return head;
}

PKGLISTNODE *DeletePkgNode(PKGLISTNODE *head, int pkgID)
{
    PKGLISTNODE *p, *q;
    if (!head)
    {
        return NULL;
    }

    if (head->pkg.id == pkgID)
    {
        p = head;
        head = head->next;
        free(p);
        return head;
    }

    q = head;
    p = head->next;
    while (p)
    {
        if (p->pkg.id == pkgID)
        {
            q->next = p->next;
            free(p);
            return head;
        }
        q = p;
        p = p->next;
    }
    return head;
}

RBTLISTNODE *DeleteRbtNode(RBTLISTNODE *head, int rbtID)
{
    RBTLISTNODE *p, *q;
    if (!head)
    {
        return NULL;
    }

    if (head->rbt.id == rbtID)
    {
        p = head;
        head = head->next;
        free(p);
        return head;
    }

    q = head;
    p = head->next;
    while (p)
    {
        if (p->rbt.id == rbtID)
        {
            q->next = p->next;
            free(p);
            return head;
        }
        q = p;
        p = p->next;
    }
    return head;
}

void FreeIntList(INTLISTNODE *head)
{
    INTLISTNODE *p, *q;
    p = head;
    while (p)
    {
        q = p->next;
        free(p);
        p = q;
    }
}


void FreePkgListNode(PKGLISTNODE *head)
{
    PKGLISTNODE *p, *q;
    p = head;
    while (p)
    {
        q = p->next;
        free(p);
        p = q;
    }
}

void FreeRbtListNode(RBTLISTNODE *head)
{
    RBTLISTNODE *p, *q;
    p = head;
    while (p)
    {
        q = p->next;
        free(p);
        p = q;
    }
}


PKGLISTNODE *FindPkgNode(PKGLISTNODE *head, int pkgID)
{
    PKGLISTNODE *p = head;
    while (p)
    {
        if (p->pkg.id == pkgID)
            return p;
        p = p->next;
    }
    return NULL;
}

INTLISTNODE *FindIntNode(INTLISTNODE *head, int val)
{
    INTLISTNODE *p = head;
    while (p)
    {
        if (p->val == val)
            return p;
        p = p->next;
    }
    return NULL;
}

RBTLISTNODE *FindRbtNode(RBTLISTNODE *head, int rbtID)
{
    RBTLISTNODE *p = head;
    while (p)
    {
        if (p->rbt.id == rbtID)
            return p;
        p = p->next;
    }
    return NULL;
}

HOMEBASENODE *NewHomeBaseNode(int x, int y)
{
    HOMEBASENODE *hb;

    hb = (HOMEBASENODE *) malloc(sizeof(HOMEBASENODE));
    if (hb == NULL)
    {
        return NULL;
    }

    hb->next = NULL;
    hb->x = x;
    hb->y = y;

    return hb;
}

HOMEBASENODE *AddHomeBaseNode(HOMEBASENODE *head, HOMEBASENODE *newNode)
{
    newNode->next = head;
    return newNode;
}

HOMEBASENODE *DeleteHomeBaseNode(HOMEBASENODE *head, HOMEBASENODE *thisNode)
{
    HOMEBASENODE *p, *prev_p;

    p = head;
    prev_p = NULL;

    while (p)
    {
        if (p == thisNode)
        {
            HOMEBASENODE *thisNext = thisNode->next;

            free(thisNode);

            if (prev_p == NULL)
            {
                return thisNext;
            }
            else
            {
                prev_p->next = thisNext;
                return prev_p;
            }
        }

        prev_p = p;
        p = p->next;
    }

    return head;
}


void IssueMove(long bid, char direction)
{
    char line[100];

    gbMyMoney -= bid;

    sprintf(line, "%ld Move %c\n", bid, direction);
    SendALine(line);
}

void IssuePick(long bid, INTLISTNODE *pkgIDs)
{
    int len;
    char defaultLine[100];
    char *line = NULL;

    gbMyMoney -= bid;

    if (pkgIDs == NULL)
    {
        sprintf(defaultLine, "%ld Pick 17\n", bid);
        SendALine(defaultLine);
        return;
    }
    else
    {
        len = IntListLen(pkgIDs);
    }

    line = (char *) malloc(4 * sizeof(char) + len * 12 + 12 + 1);
    if (line == NULL)
    {
#ifdef ICFP_DEBUG
        printf("IssuePick malloc failed\n");
#endif
    }

    if (line == NULL)
    {
        sprintf(defaultLine, "%ld Pick 17\n", bid);
        SendALine(defaultLine);
        return;
    }
    else
    {
        int i;
        char *str;

        sprintf(line, "%ld Pick", bid);

        str = line + strlen(line);

        while (pkgIDs)
        {
            sprintf(str, " %ld", pkgIDs->val);
            str += strlen(str);
            pkgIDs = pkgIDs->next;
        }

        strcpy(str, "\n");

        SendALine(line);

        if (line)
        {
            free(line);
        }

        return;
    }
}

void IssueDrop(long bid, INTLISTNODE *pkgIDs)
{
    int len;
    char defaultLine[100];
    char *line = NULL;

    gbMyMoney -= bid;

    if (pkgIDs == NULL)
    {
        sprintf(defaultLine, "%ld Pick 17\n", bid);
        SendALine(defaultLine);
        return;
    }
    else
    {
        len = IntListLen(pkgIDs);
    }

    line = (char *) malloc(4 * sizeof(char) + len * 12 + 12 + 1);
    if (line == NULL)
    {
#ifdef ICFP_DEBUG
        printf("IssueDrop malloc failed\n");
#endif
    }

    if (line == NULL)
    {
        sprintf(defaultLine, "%ld Pick 17\n", bid);
        SendALine(defaultLine);
        return;
    }
    else
    {
        int i;
        char *str;

        sprintf(line, "%ld Drop", bid);

        str = line + strlen(line);

        while (pkgIDs)
        {
            sprintf(str, " %ld", pkgIDs->val);
            str += strlen(str);
            pkgIDs = pkgIDs->next;
        }

        strcpy(str, "\n");

        SendALine(line);

        if (line)
        {
            free(line);
        }

        return;
    }
}

int main(int argc, char *argv[])
{
    PKGLISTNODE *nowPkgs = NULL;
    //Validate argv, argc

    //Connect to server
    // argv[1] contains the serverName and argv[2] the port
    DoInitPhase(argv[1], atoi(argv[2]));
    // And receive initialization info
    ReadBoard();
    ReadConfig();

#ifdef ICFP_DEBUG
    printf("ID:%d. Money:%ld. Capacity:%d.\n", gbMyID, gbMyMoney, gbMyCapacity);
#endif

    ReadSrvrResponse();

    do
    {
#ifdef ICFP_DEBUG
        printf("gbMyX = %d gbMyY = %d\n", gbMyX, gbMyY);
#endif
        nowPkgs = ReadAvailablePkgs();
        IssueRightCommand(nowPkgs);
        ReadSrvrResponse();
        FreePkgListNode(nowPkgs);

    } while (1);
}



PKGLISTNODE *ReadAvailablePkgs()
{
    PACKAGE pkg;
    PKGLISTNODE *pkgNode;
    PKGLISTNODE *nowPkgs = NULL;

    char *token;
    char *pkgList = GetNextLine();

#ifdef ICFP_DEBUG
    printf("Packagelist = :%s:\n", pkgList);
#endif
    token = strtok(pkgList, " ");
    while (token)
    {
        pkg.id = atoi(token);

        token = strtok(NULL, " ");
        pkg.destx = atoi(token);

        token = strtok(NULL, " ");
        pkg.desty = atoi(token);

        token = strtok(NULL, " ");
        pkg.weight = atoi(token);

        pkg.x = gbMyX;
        pkg.y = gbMyY;

        pkgNode = FindPkgNode(gbKnownPkgs, pkg.id);
        if (!pkgNode)
        {
            pkgNode = NewPkgNode(&pkg);
            if(pkgNode)
            {
                gbKnownPkgs = AddPkgNode(gbKnownPkgs, pkgNode);
            }
        }
        else
        {
            pkgNode->pkg = pkg;
        }

        pkgNode = NewPkgNode(&pkg);
        if(pkgNode)
        {
            nowPkgs = AddPkgNode(nowPkgs, pkgNode);
        }
        token = strtok(NULL, " ");
    }
    return nowPkgs;
}


